 /*
 
Standalone Java program to create a VRML 2.0 fractal height field surface
You need that Java Dev. Kit or some other Java Enviroment
to compile and run this program.
Changes:

10.06. changed to VRML 2.0 draft 2 syntax
30.11.96 changed to VRML 2.0 format 
		 including LOD



 Basic algorithm from 

    H. Peitgen, D. Saupe 
    "The Science of Fractal Images"
    Springer Verlag,
    Page 100



 Steps to exectue:
     javac Landscape.java
     set CLASSPATH = .;%CLASSPATH%
     java Landscape
     glview Landscape.wrl


Startet from an applet from 
C. Thornborrow (Silicon Graphics (UK) )
http://www.sgi.com/Fun/free/java/chris-thornborrow/index.html



*/
 
import java.applet.Applet;
import java.awt.Color;
import java.lang.Math;
import java.util.Random;
import java.io.*;
import java.lang.Integer;


public class Landscape extends java.applet.Applet implements Runnable 

{
    
    int      scape[][];     // height field
    Color    colors[][];    // height field colors
	float    texCoord[][][];// texture coordinates

    
    // actual amplitude  for level
    double   delta;
    
    // options 
    int      size; // size of landscape = 2^power
    int      power; // size of landscape as 2 power
    
	boolean  zeroBorder; // start with 0 at outer first level
                         // to have a landscape embedded in water 

	boolean  fixedBorder; // don't modify border 
    
	boolean  doMountain;	// compute the mountaion

	boolean  doComputeColor; // compute color values 
	boolean  doCrater;

	boolean  doPath;		// valleys
	int numPath;

	float	 maxCraterRadius;
	int		 numCraters;


    boolean  outputColor; // output a VRML file with colors (makes file much larger)
    
	long     seed;      // random seed
	int		 initHeight;
    
    

    Color    c;
    Random   rng;

    double   sigma,fracdim;
    boolean	 addition;

	String	 outFilename;

    public void defaults() {
        //size = 512;   power = 9;      sigma = 90.0;
        size = 256;   power = 8;        sigma = 64.0;  // will be 1.3 Meg 
        //size = 128;   power = 7; sigma = 64.0; // 
		setGridSize(8);
        
		initHeight = 0;

		fracdim = 0.85-0.1;
        zeroBorder = true;
        fixedBorder = false;// true; // not working yet 
        doComputeColor = false; // true;
        outputColor = true;
        addition = true;
        seed = 1;
		doMountain = true;
		doCrater = true;
		doPath = false;
		
		doPath = true;
		doMountain = false;
		doCrater = true;

		initHeight = 100;
		maxCraterRadius=10.0f;
		numCraters = 10;
		numPath = 20;

		outFilename = "Landscape.wrl";

	}
	// set the grid POWER 
	void setGridSize(int thePower)
	{
		power = thePower;
		size = 1<<power;
	}

    
    public void initialize(){
        

        if (seed != 0) rng = new Random(seed);
        else rng = new Random(); // seed based on time

        scape = new int[size+1][size+1];
        
		if (doComputeColor) 
           colors = new Color[size+1][size+1];


        int i,j;

        for (i=0; i<=size; i++)
            for (j=0; j<=size; j++) 
                scape[i][j] = initHeight;        
        
    
    }
    
    
    
    // application entry point for standalone op.
    public static void main(String argv[])
	{

        Landscape l;
        l = new Landscape();
		
		l.defaults();

		int i;
		for (i=0; i<argv.length; i++) {
			if (argv[i] == "-dim") {
				i++;
				//fracdim fromString(argv[i]);

			}
			else if (argv[i] == "-out") {
				i++;
				l.outFilename = argv[i];

			}
		}


        l.initialize();
        l.compute();

        try {
            l.save();
        } catch (java.io.IOException e) { }

    }


    public void run(){
        compute();
    }
      
    // MIdPointFM2D
    // H. Peitgen, D. Saupe The Science of Fractal Images
    // Springer Verlag Page 100

    public void compute(){
      
      int i,j,scale;
      int step,hstep;

      delta = sigma;
	 
	  if (zeroBorder)   
		  delta*=2.0f; // looks better,

	  if (true) {
      for (i=0; i<=size; i++) 
		for (j=0; j<=size; j++) {
		  scape[i][j]= initHeight;
		}

	  }

      step = size;
      hstep = step/2;

      if (!zeroBorder && ! fixedBorder) {
        // set the four corner points
        scape[0][0]      = (int) ((rng.nextGaussian())*delta);
        scape[0][size]   =(int) ((rng.nextGaussian())*delta);
        scape[size][0]   =(int) ((rng.nextGaussian())*delta);
        scape[size][size] =(int) ((rng.nextGaussian())*delta);
      }

       System.out.print("##computing Size = ");
       System.out.print(size); 
       System.out.print("## Fracdim= ");
       System.out.print(fracdim); 
       System.out.println(" ...");

        
        // create the landscape
	    if (doMountain)
        for (scale=1; scale<=power; scale++) {
        
            delta *= Math.pow (0.5, 0.5*fracdim);
			
			int start=0,end=size;
			if (fixedBorder) {
				start=step;
				end = size-step;
			}
        
            // set the internal point in square
            for (i=hstep; i<=size-hstep; i+=step)
                for (j=hstep; j<=size-hstep; j+=step)
			//	this.set4pointx(i,j,step,hstep);
				this.set4point (i,j, scale, 
                                    scape[i+hstep][j+hstep],
                                    scape[i+hstep][j-hstep],
                                    scape[i-hstep][j+hstep],
                                    scape[i-hstep][j-hstep]);
            
  // displace other points
            if (addition)
                for (i=start; i<end; i+=step)
                    for (j=start; j<=end; j+=step)
                        scape[i][j] +=(int) ((rng.nextGaussian())*delta);


                                    
            delta *= Math.pow (0.5, 0.5*fracdim);
                                    
            // set boundary points
            for (i=hstep; i<=size-hstep; i+=step) {
                this.set3point (i, 0, scale-1,scape[i+hstep][0],
                                scape[i-hstep][0], scape[i][hstep]);
                this.set3point (i,size,scale -1, scape[i+hstep][size],
                                scape[i-hstep][size], scape[i][size-hstep]);
                this.set3point (0,i, scale -1, scape[0][i+hstep],
                                scape[0][i-hstep],scape[hstep][i]);
                this.set3point (size,i, scale-1, scape[size][i+hstep],
                                scape[size][i-hstep],scape[size-hstep][i]);
            }
            
            // set interior grid points
            for (i=hstep; i<=size-hstep; i+=step)
                for (j=step; j<=size-hstep; j+=step)
//					this.set4pointx(i,j,step,hstep);
                  this.set4point (i,j,scale-1, 
                                    scape[i][j+hstep], scape[i][j-hstep],
                                    scape[i+hstep][j], scape[i-hstep][j]);
            for (i=step; i<=size-hstep; i+=step)
                for (j=hstep; j<=size-hstep; j+=step)
//					this.set4pointx(i,j,step,hstep);
                   this.set4point (i,j, scale-1,
                                   scape[i][j+hstep], scape[i][j-hstep],
                                   scape[i+hstep][j],scape[i-hstep][j] ); 
            // displace other points
            if (addition) {
                for (i=start; i<=end; i+=step)
                    for (j=start; j<=end; j+=step)
                        scape[i][j] +=(int) ((rng.nextGaussian())*delta);

                for (i=hstep; i<=size-hstep; i+=step)
                    for (j=hstep; j<=size-hstep; j+=step)
                        scape[i][j] +=(int) ((rng.nextGaussian())*delta);

            }

            if (zeroBorder && !fixedBorder) 
                for (i=0; i<=size; i+=step)
                {
                    scape[i][0]=0;
                    scape[i][size]=0;
                    scape[0][i]=0;
                    scape[size][i]=0;
                }

                            
            // change the size of the grid we are working on
            step /=2;
            hstep /=2;
        }

		if (doCrater)
			//for (i=0; i<size;i+= 4) crater();
		    for (i=0; i<numCraters;i++) crater(); 


		if (doPath)
		    //for (i=0; i<size;i+= 6) 
		    for (i=0; i<numPath;i++) 
			  path();

 		if (doComputeColor) {
         System.out.println("computing colors ..");
        
         for (i=0; i<=size; i++) 
              for (j=0;  j<=size; j++) {
                    computeColor(i,j,scape[i][j]);

                    if (scape[i][j] < 0) scape[i][j] = (int) (scape[i][j]*0.1);

               }
        }

        System.out.println("computing done ..");



                    
    }

    private final void set4point (int i, int j, int s, int a, int b, int c, int d) {
        
        scape[i][j] =   ((a + b +c +d )/4) +  (int)(rng.nextGaussian() * delta);
    }
    
    private final void set3point (int i, int j, int s, int a, int b, int c) {
    
        scape[i][j] = ((a + b + c )/3) + (int)(rng.nextGaussian() * delta);
    }

    private final void set4pointx (int i, int j, int step, int hstep) {
		int first = scape[i][j+hstep]+ scape[i][j-hstep] +
                                    scape[i+hstep][j]+ scape[i-hstep][j];

		int second = 0;
		int  dx = hstep+step;
		
		if (((j-dx) >=0 && (j+dx) <=size) && ((i-dx) >=0 && (i+dx) <=size) ) {
			second += scape[i][j+dx]+ scape[i][j-dx];
            second += scape[i+dx][j]+ scape[i-dx][j];
			first = (6 * first + 4 * second) / (4*10);
		} else
			first /=4;


        scape[i][j] =   (first) +  (int)(rng.nextGaussian() * delta);
    }

	void crater(int i,int j, int r, int h0,int h1) {
		int imin = i-r; if (imin<0) imin = 0;
		int imax = i+r;	if (imax>size) imax = size;
		int jmin = j-r; if (jmin<0) jmin = 0;
		int jmax = j+r;	if (jmax>size) jmax = size;
		if (r<=0) return;

		int a,b;
		int r2=r*2;
		for(a=imin; a<=imax;a++) { 
			int da=a-i;da=da*da;
			for(b=jmin; b<=jmax;b++) {
				int db=b-j;db=db*db;
				int d = da+db;
				if (d >= r2) continue;
				float h= (h0 + (h1-h0)) * (float)d/(float)r2;
				if (scape[a][b] >= h)
				    scape[a][b]= (int) h;
			}
		}	

	}
	
	public void crater() {
		int i=(int) (rng.nextFloat()*size);
		int j=(int) (rng.nextFloat()*size);
		int r=(int) (rng.nextFloat()*maxCraterRadius*(float)power);
		crater(i,j,r,-10,80);
	}

	// compute a valley by a randomized line 
	void path(int i1,int j1, int i2,int j2, int r, int h0,int h1) {
		
		int di = (i2-i1);
		if (di<0) di = -di;
		int dj = (j2-j1);
		if (dj<0) dj = -dj;
		int i = (i1+i2) / 2;
		int j = (j1+j2) / 2;
		int dist= di+dj;
		if ( ((di >2) || (dj >2)) && (dist > (r-3))) { // subdivide 

		   i +=  (int) (((float) rng.nextGaussian()) * dist *0.2);
		   j +=  (int) (((float) rng.nextGaussian()) * dist *0.2);
		   
		   path(i1,j1,i,j,r,h0,h1);
		   path(i,j,i2,j2,r,h0,h1);

		   return;
		}

		
		int imin = i-r; if (imin<0) imin = 0;
		int imax = i+r;	if (imax>size) imax = size;
		int jmin = j-r; if (jmin<0) jmin = 0;
		int jmax = j+r;	if (jmax>size) jmax = size;
		if (r<=0) return;

		int a,b;
		int r2=r*2;
		for(a=imin; a<=imax;a++) { 
			int da=a-i;da=da*da;
			for(b=jmin; b<=jmax;b++) {
				int db=b-j;db=db*db;
				int d = da+db;
				if (d >= r2) continue;
				//float h= (h0 + (h1-h0)) * (float)d/(float)r2;
				//if (scape[a][b] >= h)
				int h=h0;
				    scape[a][b]= (int) h;
			}
		}	

	}
	
	public void path() {
		int i1=(int) (rng.nextFloat()*size);
		int j1=(int) (rng.nextFloat()*size);
		int i2=(int) (rng.nextFloat()*size);
		int j2=(int) (rng.nextFloat()*size);
		int r=(int) ( 1+ rng.nextFloat()*0.2*(float)power);
		
		//r=1;
		path(i1,j1,i2,j2,r,-10,80);
	}

	// compute max height 
	public int getMaxHeight() 
	{

		 int height = scape[0][0];
		 int i,j;
         for (i=0; i<=size; i++) 
              for (j=0;  j<=size; j++) {
                    if (scape[i][j] > height) height = scape[i][j];
		 }

		return height;
	}
	// compute texture coordinates
	public void computeTexCoords() 
	{

		 int maxHeight = getMaxHeight();
		 
		 texCoord = new float [size+1][size+1][2];


		 int i,j;
         for (i=0; i<=size; i++) {
              for (j=0;  j<=size; j++) {
				    float h = (float) scape[i][j] / (float) maxHeight;
					
					h += ((float) rng.nextGaussian()) * 0.05f; // add some randomness 

					if (h<0) h= 0.0f;

					texCoord[i][j][0]= (i*j) * 0.001f;
					texCoord[i][j][1]= h;
			} 
		 }	
	}

    // save Heighfield as VRML file 
    public void save()
            throws  java.io.IOException
    {
            FileOutputStream outFile;
            PrintStream data;
            int i,j;

            System.out.print("saving ");
            System.out.print(outFilename);
            System.out.println(" ...");
            
            // first open the file
            
            outFile = new FileOutputStream(outFilename);
            // create a data stream
            data   = new PrintStream( outFile );

            data.println("#VRML V2.0 utf8");
            data.println("#heightfield demo created by Landscape.java V 0");
			data.println("NavigationInfo { headlight TRUE 	type \"WALK\" avatarSize 1.0 }");
			if (true) {
			data.println("Background {"); 
			data.println("groundAngle  1.5707958 	groundColor  [ 0.5  0.0 0.9 , 0 0.0 1.0 ]"); 
			data.println("skyAngle  1.5707958 skyColor  [ 0.46 0.38 0.88, 0.8 0.7 0.94 ]"); 
			data.println("}"); 
			}
			if (false) {
			data.println("DirectionalLight {");
			data.println("	color  1.0000 1.0000 1.0000"); 
			data.println("	direction  -1.0 1. 1.0");
			data.println("	intensity  1.0000");
			data.println("	on  TRUE");
			data.println("}");
			}



            data.println("Group {");
            data.println("children [");

			if (true) {
				int l = 2;
				//saveSubGrid4(data,0,size,0,size,1,1, outputColor);
				saveLodGrid(data,l,0,size,0,size,1<<l,1<<l, outputColor);
//				saveLodGrid(data,l,0,size,0,size,16,16, outputColor);
			}
			else {
            data.println("Shape {");
            data.println("appearance Appearance {");

            data.println("}");

            data.println("geometry ");
            double f = 1.0/ 255.0;
            


            // VRML convention Y is Z
            f = 0.15;
            //data.println("Scale { scaleFactor 1.0 "); data.print(f); data.println(" 1.0}"); 


            // ********** output the grid
            data.println("ElevationGrid {");
                data.print("xDimension "); data.print(size+1); data.println();
                data.print("zDimension "); data.print(size+1); data.println();
                
				//data.print("xSpacing "); data.print(1.0/(size+1)); data.println();
                //data.print("zSpacing "); data.print(1.0/(size+1)); data.println();
				data.print("xSpacing "); data.print(1.0); data.println();
                data.print("zSpacing "); data.print(1.0); data.println();
                
				data.print("#fracdim "); data.print(fracdim); data.println();

                data.println("solid TRUE"); 
                data.println("ccw FALSE"); 
                data.println("height [ "); 
            
				for (i=0; i<=size; i++) {
					for (j=size; j>=0; j--) {
						data.print(scape[i][j]*f);
						data.print(",");
					}
					data.println();
				}
				data.println("]");  // height 
				// ********* write per vertex color 
				if (outputColor) {

				data.println("color Color {");
					data.println("color [ "); 
            

				for (i=0; i<=size; i++) {
					for (j=size; j>=0; j--) {
						Color c= colors[i][j];
						data.print(c.getRed() *f); data.print(" ");
						data.print(c.getGreen() *f); data.print(" ");
						data.print(c.getBlue() *f); data.print(" ");
						data.print(",");
					}
					data.println();
				}
            
				data.println("]"); 
				data.println("}");

				}




            data.println("}"); // ElevationGrid 
            data.println("}"); // Shape
			
			}
            data.println("]}"); // Group

            outFile.close();
    }


	// save a sub  Heighfield to VRML file
    public void saveSubGrid4(PrintStream data, 
			int xmin,int xmax,int ymin,int ymax,int xstep,int ystep,boolean doColor)
            throws  java.io.IOException
    {
            int i,j;

			int xsize = (xmax-xmin+1);
			int ysize = (ymax-ymin+1);

			int xc = xmin + xsize / 2;
			int yc = ymin + ysize / 2;
			
			saveSubGrid(data,xmin,xc,ymin,yc,1,1,doColor);
			
			saveSubGrid(data,xc,xmax,ymin,yc,1,1,doColor);
			
			saveSubGrid(data,xmin,xc,yc,ymax,1,1,doColor);
			
			saveSubGrid(data,xc,xmax,yc,ymax,1,1,doColor);

	}

	// save a sub  Heighfield to VRML file
    public void saveLodGrid(PrintStream data, 
			int level,
			int xmin,int xmax,int ymin,int ymax,int xstep,int ystep,boolean doColor)
            throws  java.io.IOException
    {

			int xsize = (xmax-xmin+1);
			int ysize = (ymax-ymin+1);
			int xc = xmin + xsize / 2;
			int yc = ymin + ysize / 2;

			if (level <=1|| xstep == 1 || ystep == 1) {
				saveSubGrid(data,xmin,xmax,ymin,ymax,1,1,doColor);
				return;

			}


            data.println("LOD {");
            data.print("center "); data.print(xmin+xsize/2.0); data.print(" 0 "); data.print(ymin+ysize/2.0); data.println();
			data.print("range ["); 
				data.print(xsize);  
			data.println("]");
			
			data.print("level ["); 

			

			// lod 1
			// inline 
            data.println("Group {");
				data.println("children [");
			
					saveLodGrid(data,level-1,xmin,xc,ymin,yc, xstep/2,ystep/2,doColor);
			
					saveLodGrid(data,level-1,xc,xmax,ymin,yc,xstep/2,ystep/2,doColor);
			
					saveLodGrid(data,level-1,xmin,xc,yc,ymax,xstep/2,ystep/2,doColor);
			
					saveLodGrid(data,level-1,xc,xmax,yc,ymax,xstep/2,ystep/2,doColor);

            data.println("]}");

			saveSubGrid(data,xmin,xmax,ymin,ymax,xstep,ystep,doColor);

            data.println("]}"); //


	}




    
	// save a sub  Heighfield to VRML file
    public void saveSubGrid(PrintStream data, 
			int xmin,int xmax,int ymin,int ymax,int xstep,int ystep,boolean doColor)
            throws  java.io.IOException
    {
            int i,j;

			int xsize = (xmax-xmin+1);
			int ysize = (ymax-ymin+1);


            System.out.print("saving grid  ");
            System.out.println(" ...");
            
            // create a transform 
            
            data.println("Transform {");
            data.print("translation "); data.print(xmin); data.print(" 0 "); data.print(ymin); data.println();

            data.println("children [");

            data.println("Shape {");
            data.println("appearance Appearance {");

            data.println("}");

            data.println("geometry ");
            double f = 1.0/ 255.0;
            

            // VRML convention Y is Z
            f = 0.15;
            //data.println("Scale { scaleFactor 1.0 "); data.print(f); data.println(" 1.0}"); 


            // ********** output the grid
            data.println("ElevationGrid {");
                data.print("xDimension "); data.print(1+xsize/xstep); data.println();
                data.print("zDimension "); data.print(1+ysize/ystep); data.println();
				
				data.print("xSpacing "); data.print(xstep); data.println();
                data.print("zSpacing "); data.print(ystep); data.println();

             

                data.println("solid TRUE"); 
                data.println("ccw FALSE"); 

                data.println("height [ "); 
            
				for (i=ymin; i<=ymax; i+=ystep) {
					int cnt=0;
					for (j=xmin; j<=xmax; j+=xstep) {
						data.print(scape[i][j]*f);
						data.print(",");
						cnt++;
						if (cnt%64 == 1) data.println(); // otherwise crash
					}
					data.println();
				}
				data.println("]");  // height 
				
				// ********* write per vertex color 
				if (outputColor) {

				data.println("color Color {");
					data.println("color [ "); 
            
				for (i=ymin; i<=ymax; i+=ystep) {
					int cnt=0;
					for (j=xmin; j<=xmax; j+=xstep) {
						Color c= colors[i][j];
						data.print(c.getRed() *f); data.print(" ");
						data.print(c.getGreen() *f); data.print(" ");
						data.print(c.getBlue() *f); data.print(" ");
						data.print(",");
						cnt++;
						if (cnt%32 == 1) data.println(); // otherwise crash

					}
					data.println();
				}
            
				data.println("]"); 
				data.println("}");

				}




            data.println("}"); // ElevationGrid 
            data.println("}"); // Shape
            data.println("]}"); // Group
    }

    
    
    // compute a color for grid point i,j based on h
    private final void computeColor (int i, int j, int h) {
		int red,green,blue;
    
        if (j==0) 
            c = (Color.black);
        else if (h<=0) {
			blue = 200+h;
            if (blue>255) blue = 255; else if (blue <0) blue = 0;
            c = (new Color (0,0,blue));
        } else if (h<5)
            c = (new Color (218,195,112));
        else if (h<110) {
            red = 0;
            green = 50+h/2;
            if (h>30) red = h-10;
            c = (new Color(red,green,0));
        }
        else if (h>=110) {
            red = h+100;
            if (red>255) red = 255;
            c = (new Color(red-10,red-10,red));
        }
           
        colors[i][j]= c;

    }
         
}
