//
//
//  VRML 2.0 EAI fractal Landscape sample
//
// tested with blaxxunCC3D 
// extended version with BspTree construction
// surface features etc.
//
// holger@blaxxun.com

import java.applet.*;
import java.awt.*;
import java.util.*;

import vrml.external.field.*;
import vrml.external.exception.*;
import vrml.external.Node;
import vrml.external.Browser;



import Landscape;



/*
 *
 * LandscapeApp
 *
 */
public class LandscapeApp extends Applet
{

  // Browser we're using
  vrml.external.Browser browser;
  boolean blaxxunCC3D; // are we running blaxxunCC3D ? 

  // Root of the scene graph 
  Node root;

  Node worldInfo;	 // some nodes of interest in root.wrl
  Node navigationInfo;
  Node viewpoint;

  // the landscape we are constructing

  Landscape s;

  Node grid; // elevationGrid
  Node gridTexCoord; // elevationGrid TexCoord

  int gridSize; // current grid size
  float xSpacing; // spacing between grid points (m)
  float zSpacing;
  
  float xOffset;	// offset of origin 
  float zOffset;

  float heightScale;		// y (height) scale factor
  float heightOverGround;

  boolean saveWorld; // true save resulting VRML scene 

  String outputPath;

  boolean buildTiles; // subdivide into tiles
  boolean buildBsp; // build blaxxunCC3D BSP-Tree 
  boolean buildInlines; // build tiles and split into inlines 

  boolean doTextured;	// create texture coordinates
  boolean doColored;	// create colorPerVertex
  
  String  landscapeTexture; // URL for landscapeTexture y = height
  
  Node landscapeAppearance; // material used for landscape appearance

 


  // called when the browser handle is available
  public void setBrowser(Object theBrowser)
  {
	System.out.println("setBrowser called " );
	if (theBrowser instanceof vrml.external.Browser) {
		browser = ( vrml.external.Browser)  theBrowser;
	} else {
		// create browser object from some handle 
		browser =  vrml.external.Browser.getBrowser(theBrowser);
	}

	if (browser == null) {
	  showStatus("Can't get browser");
      System.out.println("Can't get browser");
	  return;
	}		
  

	blaxxunCC3D = browser.getName().toLowerCase().compareTo("blaxxuncc3d") == 0;
	if (blaxxunCC3D)
		showStatus("Connected to blaxxunCC3D");
	else showStatus("Connected to browser");

	getRootNode();

  
  }

  // Get handle to root of the scene graph
  public boolean getRootNode()
  {

	try {
      root = browser.getNode("ROOT");
    }						  
    catch (InvalidNodeException e) {
	  //showStatus("No Root node yet in WRL File");
      //System.out.println("No Root node in WRL File: " + e);
	  root = null;
    }
	
	try { navigationInfo = browser.getNode("NavigationInfo"); } catch (Exception e) { }
	try { worldInfo = browser.getNode("WorldInfo"); } catch (Exception e) { }
	try { viewpoint = browser.getNode("Viewpoint"); } catch (Exception e) { }

	return root != null;
  }


  public Browser getBrowser() {
    return browser;
  }


  // create a VRML node 
  public Node createNode(String nodeType) {
	  Node node[] = browser.createVrmlFromString(nodeType + " { } ");
	  return node[0];
  }	


  public void clearScene() {
  		Node [] children;
		children = ((EventOutMFNode) root.getEventOut("children_changed")).getValue();
		
		((EventInMFNode) root.getEventIn("removeChildren")).setValue(children);
		grid = null;
		gridTexCoord = null;
  }	



	String  pinfo[][]  =  {
	   {"fracDim",		"float",		"float 0 .. 1 fractalness"},
	   {"seed",			"int",			"random seed 1.... 0  use time based seed"}
	};

    // init 

	public String[][] getParameterInfo() { return pinfo; }

  
  public void init() {
		
		
		//default values 
		fracDim = 0.25f; // actuall 2.5 2 -- 3 
		seed = 1; // 
		power = 6;
		sigma = 64.0f;

		xSpacing = 1.0f;
		zSpacing = 1.0f;
	    heightScale	= 0.05f;
		heightOverGround=1.85f; // from nav info 
		
		outputPath = "d:\\landscape\\";
		landscapeTexture = "mount3.jpg";

		buildBsp = true;

		//buildBsp = false;
		buildTiles = buildBsp;

		saveWorld = false;
		//saveWorld = true;
		doTextured = true;
		doColored = false;


		// create our widgets 
		makeUI();

		// try static get browser method  (Netscape)
		Browser b= Browser.getBrowser(this);
		if (b != null) {
			setBrowser(b);
		}

  }	

  public String getParamsString() {
	  String s;
	  s="";
	  s+= "size="+String.valueOf(gridSize);
	  s+= " fracDim="+String.valueOf(fracDim);
	  s+= " seed="+String.valueOf(seed);
	  s+= " sigma="+String.valueOf(sigma);
   	  s+= " numValleys="+String.valueOf(numValleys);
   	  s+= " numCraters="+String.valueOf(numCraters);
   	  s+= " xSpacing="+String.valueOf(xSpacing);
   	  s+= " zSpacing="+String.valueOf(zSpacing);
   	  s+= " heightScale="+String.valueOf(heightScale);

	  return s;
  }


  public void start() {
  }	

  public void stop() {
  }	


  // UI Stuff 

  float			fracDim;	 // fractalness 
  Scrollbar		fracDimScr;
  TextField		fracDimField;

  int			seed;		// random seed, 0 use time
  Scrollbar		seedScr;
  TextField		seedField;

  int			power;		// grid power
  TextField		powerField;

  float			sigma;	 // fractalness 
  Scrollbar		sigmaScr;
  TextField		sigmaField;

  Scrollbar		spacingScr;
  TextField		spacingField;

  int			numValleys;	
  TextField		valleysField;

  int			numCraters;
  TextField		cratersField;

  // UI helpers

  public void makeButton(String name,
			 GridBagLayout gridBag,
			 GridBagConstraints c) {
    Button button = new Button(name);
    gridBag.setConstraints(button, c);
    add(button);
  }

  public void makeLabel(String name,
			GridBagLayout gridBag,
			GridBagConstraints c) {
    Label label = new Label(name, Label.CENTER);
    gridBag.setConstraints(label, c);
    add(label);
  }

  public Scrollbar makeScrollbar(GridBagLayout gridBag,
				 GridBagConstraints c,
				 int orientation,
				 int value,
				 int min,
				 int max) {
    Scrollbar bar = new Scrollbar(orientation, value, 1, min, max);
    gridBag.setConstraints(bar, c);
    add(bar);
    return bar;
  }


  public void makeUI() {

  
    GridBagLayout gridBag = new GridBagLayout();
    
	GridBagConstraints c = new GridBagConstraints();
   
	setFont(new Font("Helvetica", Font.PLAIN, 12));
    setLayout(gridBag);

	// parameter fracDim
    c.fill = GridBagConstraints.BOTH;

	c.gridx = 0;
	c.gridy = 0;
    c.gridwidth = 1; c.gridheight = 2;

	c.weightx = 0.1;

    makeLabel("Fractal Dimension", gridBag, c);

    c.gridwidth = 6;  c.gridheight = 2;
	c.gridx = GridBagConstraints.RELATIVE;

    fracDimScr = makeScrollbar(gridBag, c, Scrollbar.HORIZONTAL, (int) (fracDim*1000f), 0, 1000);

    c.gridwidth = 2;  c.gridheight = 2;
	
    fracDimField = new TextField(Float.toString(fracDim));
    gridBag.setConstraints(fracDimField, c);
    add(fracDimField);

	// parameter sigma
    c.fill = GridBagConstraints.BOTH;
	c.gridx = 0;
	c.gridy = 2;
    c.gridwidth = 1; c.gridheight = 2;

	c.weightx = 0.1;

    makeLabel("Sigma", gridBag, c);

    c.gridwidth = 6;  c.gridheight = 2;
	c.gridx = GridBagConstraints.RELATIVE;
	//c.gridy = GridBagConstraints.RELATIVE;

    sigmaScr = makeScrollbar(gridBag, c, Scrollbar.HORIZONTAL, (int) (sigma*10f), 0, 10000);
    c.gridwidth = 2;  c.gridheight = 2;
	
    sigmaField = new TextField(String.valueOf(sigma));
    gridBag.setConstraints(sigmaField, c);
    add(sigmaField);


	// numValley
    makeLabel("#Valleys", gridBag, c);
    valleysField = new TextField(String.valueOf(numValleys));
    gridBag.setConstraints(valleysField, c);
    add(valleysField);


	// parameter seed
    c.fill = GridBagConstraints.BOTH;
	c.gridx = 0;
	c.gridy = 4;
    c.gridwidth = 1; c.gridheight = 2;

	c.weightx = 0.1;

    makeLabel("Seed", gridBag, c);

    c.gridwidth = 6;  c.gridheight = 2;
	c.gridx = GridBagConstraints.RELATIVE;
	//c.gridy = GridBagConstraints.RELATIVE;

    seedScr = makeScrollbar(gridBag, c, Scrollbar.HORIZONTAL, seed, 0, 10000);

    c.gridwidth = 2;  c.gridheight = 2;
	
    seedField = new TextField(String.valueOf(seed));
    gridBag.setConstraints(seedField, c);
    add(seedField);

	// numCraters
    makeLabel("#Craters", gridBag, c);
    cratersField = new TextField(String.valueOf(numCraters));
    gridBag.setConstraints(cratersField, c);
    add(cratersField);


	// parameter spacing
    c.fill = GridBagConstraints.BOTH;
	c.gridx = 0;
	c.gridy = 6;
    c.gridwidth = 1; c.gridheight = 2;

	c.weightx = 0.1;

    makeLabel("Spacing", gridBag, c);

    c.gridwidth = 6;  c.gridheight = 2;
	c.gridx = GridBagConstraints.RELATIVE;
	//c.gridy = GridBagConstraints.RELATIVE;

    spacingScr = makeScrollbar(gridBag, c, Scrollbar.HORIZONTAL, (int) (xSpacing*10.0f), 0, 1000);

    c.gridwidth = 2;  c.gridheight = 2;

    spacingField = new TextField(String.valueOf(xSpacing));
    gridBag.setConstraints(spacingField, c);
    add(spacingField);



	// power 
    makeLabel("Grid Power", gridBag, c);
    powerField = new TextField(String.valueOf(power));
    gridBag.setConstraints(powerField, c);
    add(powerField);



	c.gridx = 2;
	c.gridy = 8;
    c.gridwidth = 6;
    c.gridheight = 4;

    makeButton("Go", gridBag, c);


  }	
	
  // compute surface, and update/create VRML ElevationGrid

	public boolean OnCompute()
	{

		if (browser == null) {
			showStatus("Error : not connected to browser");
			return false;
		}

		if (root == null) {
			if (!getRootNode()) {
				showStatus("Can't get root node yet, please retry");
				return false;
			}
		}


		
		if (s == null) {
			s = new Landscape();
			s.defaults();
		}

		s.fracdim = 1.0-fracDim;
		s.doComputeColor = doColored;
		s.zeroBorder = true;
		s.seed = seed;
		s.sigma = sigma;
		
		s.setGridSize(power);
		
		s.numPath = numValleys;
		s.doPath = numValleys >0;
		s.numCraters = numCraters;
		s.doCrater = numCraters >0;

		s.doMountain = sigma != 0;

		s.initialize();
		
		showStatus("Computing Surface "+(s.size)+" * " +(s.size)+" ...");
        
		s.compute();

		if (doTextured) {
			showStatus("Computing Texture coordinates "+(s.size)+" * " +(s.size)+" ...");
			s.computeTexCoords();
		}


		showStatus("Surface computed ");
		
		browser.beginUpdate();
		{

		if (landscapeAppearance == null) {
			initializeAppearance();
		}		

		int size = s.size; // the grid size 

		showStatus("Computing VRML Grid "+(s.size)+" * " +(s.size)+" ...");

		if (buildTiles) {
			int level = 4; // termination also if grid size <=17
			clearScene();

			Node n[] = new Node [1];
			n[0] = makeGrid(0,size,0,size,0,level);

			EventInMFNode addChildren = 
				(EventInMFNode) root.getEventIn("addChildren");
			addChildren.setValue(n);

		} else {

		if ((size+1)!=gridSize && grid !=null) // dimension changed //  && !blaxxunCC3D)
		    clearScene(); 

		if (grid == null) {
			Node [] n;

			//create our small scene graph
			n = browser.createVrmlFromString("ElevationGrid { creaseAngle 1.5 xDimension "+(size+1)+" zDimension "+(size+1)+" xSpacing "+String.valueOf(xSpacing) + " zSpacing "+String.valueOf(zSpacing)+ " }");
			grid = n[0];
			gridSize = size +1;
			
			if (blaxxunCC3D) grid.setName("grid");			

			n[0] = browser.createNode("Shape");

			((EventInSFNode) n[0].getEventIn("set_appearance")).setValue(landscapeAppearance);
			((EventInSFNode) n[0].getEventIn("set_geometry")).setValue(grid);

			EventInMFNode addChildren = 
				(EventInMFNode) root.getEventIn("addChildren");
			addChildren.setValue(n);

		}

		showStatus("Displaying Surface ...");
		
		// fill elevationGrid with current values 
/*
		// according to VRML spec this fields are not exposed, sorry, but in blaxxunCC3D 
		if (blaxxunCC3D) {
			((EventInSFInt32) grid.getEventIn("set_xDimension")).setValue(size+1);
			((EventInSFInt32) grid.getEventIn("set_zDimension")).setValue(size+1);
			((EventInSFFloat) grid.getEventIn("set_xSpacing")).setValue(100f/(float)size);
			((EventInSFFloat) grid.getEventIn("set_zSpacing")).setValue(100f/(float)size);
			((EventInSFFloat) grid.getEventIn("set_creaseAngle")).setValue(1.5);
		}
*/

		// make flat MFFloat array
		float heights[] = new float[(size+1) * (size+1)];

		float tc[][];
		//if (doTextured)
		tc = new float [(size+1) * (size+1)][2];

		int i,j;
		float f = heightScale; // height scale factor from int
		int out=0;
		int surf[][]=s.scape;
				for (i=0; i<=size; i++) {
					for (j=size; j>=0; j--) {
						heights[out] = surf[i][j]*f;
						if (doTextured) {
							tc[out][0] = this.s.texCoord[i][j][0];
							tc[out][1] = this.s.texCoord[i][j][1];
						}

						out ++;
					}
				}


		showStatus("setting height in vrml browser .... ");

		((EventInMFFloat) grid.getEventIn("set_height")).setValue(heights);

		heights = null;

		// tex coord 
		if (doTextured) {
			Node texCoord;
			texCoord = makeTexCoord(tc);
			tc = null;
			if (texCoord !=null) 
				((EventInSFNode)  grid.getEventIn("set_texCoord")).setValue(texCoord);
			texCoord = null;
		}

		
		}

		if (viewpoint != null) {
			float vp[] = new float[3];

			int x = size / 2;
			int y = 1;
			// to do: compute a new viewpont 
			vp[0] = getX(x);
			vp[1] = getHeight(x,y)+heightOverGround;
			vp[2] = getZ(y);
			((EventInSFVec3f) viewpoint.getEventIn("set_position")).setValue(vp);
		}

		if (worldInfo != null) {
		   ((EventInMFString) worldInfo.getEventIn("set_info")).set1Value(0,getParamsString());
		}


		if (blaxxunCC3D && saveWorld) {
			String fileName;
			fileName = outputPath +"landscape";
			fileName +="."+size;
			if (buildBsp) fileName += ".bsp";
			fileName += ".wrl";
			showStatus("saving VRML file "+fileName+" ...");
		    showStatus("Saving VRML Scene "+(s.size)+" * " +(s.size)+" to "+fileName+" ...");

			browser.saveWorld(fileName);
		}
		} 
		browser.endUpdate();

		showStatus("Surface done");


		return true;

	}

	void initializeAppearance() {

			Node a=createNode("Appearance");
			Node m=createNode("Material");

			float c[] = new float[3];
			c[0]= 0.5f; c[1]= 0.7f; c[2]= 0.4f;
			((EventInSFColor) m.getEventIn("set_diffuseColor")).setValue(c);

			c[0]= 1.0f; c[1]= 1.0f; c[2]= 1.0f;
			((EventInSFColor) m.getEventIn("set_specularColor")).setValue(c);

			((EventInSFNode) a.getEventIn("set_material")).setValue(m);

			if (doTextured) {
				Node t= createNode("ImageTexture");

				((EventInMFString) t.getEventIn("set_url")).set1Value(0,landscapeTexture);

				((EventInSFNode) a.getEventIn("set_texture")).setValue(t);

			}
			a.setName("landscapeApp");

			landscapeAppearance = a;
			m=null;
	}
	
	float getX(int x) { return (float)x*xSpacing; }
	float getZ(int z) { return (float)z*zSpacing; }

	float getHeight(int x,int y) {
		return s.scape[y][x]*heightScale;
	}

	// save a subrectangle of surface 
	Node createGridNode(int x1,int x2,int y1,int y2, float spacing, int scape[][]) 
	{
		    Node shape;
			Node grid;

			float s = spacing;
			int sizeX = (x2-x1+1);
			int sizeY = (y2-y1+1);

			
			if (s == 0.0f) s = 1.0f / (float) (sizeX);

			{
			//create our small scene graph
			Node [] n;
			n = browser.createVrmlFromString("ElevationGrid {creaseAngle 1.5  xDimension "+(sizeX)+" zDimension "+(sizeY)+
				" xSpacing "+String.valueOf(s) +" zSpacing "+String.valueOf(s) + "}");
			grid = n[0];
		
			shape = browser.createNode("Shape");

			((EventInSFNode) shape.getEventIn("set_appearance")).setValue(landscapeAppearance);
			((EventInSFNode) shape.getEventIn("set_geometry")).setValue(grid);


		}

		// make flat MFFloat array
		float heights[] = new float[(sizeX) * (sizeY)];
		float tc[][];
		//if (doTextured)
		tc = new float [sizeX * (sizeY)][2];


		int i,j;
		float f = heightScale; // height scale factor 
		int out=0;
			for (i=y1; i<=y2; i++) {
				for (j=x1; j<=x2; j++) {
					float h = scape[i][j]*f;
					
					heights[out] = h;
					if (doTextured) {
						tc[out][0] = this.s.texCoord[i][j][0];
						tc[out][1] = this.s.texCoord[i][j][1];
					}
/*		
					tc[out][0]= (i*j) * 0.001f;
					
					h += ((float) this.s.rng.nextGaussian()) * 0.2f; // add some randomness 
					h *= 0.2f;
					if (h<0) h= 0.0f;
					tc[out][1]= h;
*/
					out++;
				}
			}

		// heights 
		((EventInMFFloat) grid.getEventIn("set_height")).setValue(heights);
		heights = null;


		// tex coord 
		if (doTextured) {
			Node texCoord;
			texCoord = makeTexCoord(tc);
			tc = null;
			if (texCoord !=null) 
				((EventInSFNode)  grid.getEventIn("set_texCoord")).setValue(texCoord);
			texCoord = null;
		}

		//  color 

		Node n;

		n = createNode("Transform");
		{
		//t[0] = 3.0f; t[1] = 3.0f; t[2] = 3.0f;
		//((EventInSFVec3f) n.getEventIn("set_scale")).setValue(t);
		float t[] = new float[3];

		t[0] =x1 * xSpacing; t[1] = 0; t[2] = y1 * zSpacing;
		((EventInSFVec3f) n.getEventIn("set_translation")).setValue(t);

		}

		((EventInMFNode) n.getEventIn("addChildren")).set1Value(0,shape);

		return  n;

	}
	
	Node makeTile(int x1,int x2,int y1,int y2)
	{
		Node n;
		n = createGridNode(x1,x2,y1,y2,xSpacing,s.scape);
		return n;

	}

	// create a node representing a tree node 
	Node makeNode(Node left,Node right,int splitDim,float splitOffset) 
	{

		Node g;
		if (buildBsp) { // use blaxxunCC3D BspTree node 
			g = createNode("BspTree");
			((EventInSFNode) g.getEventIn("set_front")).setValue(left);
			((EventInSFNode) g.getEventIn("set_back")).setValue(right);
			float plane[]= new float[4];
			plane[0]=plane[1]=plane[2]=0.0f;
			if (splitDim == 0) plane[0]= 1.0f;
			else plane[2]= 1.0f;
			// D = - (Ax +By + Cz)
			plane[3]= -splitOffset;
			((EventInSFRotation) g.getEventIn("set_plane")).setValue(plane);

		} else {
			g = createNode("Group");
			EventInMFNode addChildren = 
				(EventInMFNode) g.getEventIn("addChildren");

			addChildren.set1Value(0,left);
			addChildren.set1Value(0,right);
		}

		return(g);

	}


	// recursively split rectangualar area in tree 
	// split=0 split along x, = 1 : split along y
	// level = max split level
	Node makeGrid(int x1,int x2,int y1,int y2,int split,int level)
	{
		Node n;
		if ((level <=0) || ((split == 0) && ((x2-x1)<=17))) 
			return makeTile(x1,x2,y1,y2);

		if (split == 0) { // split along x

			int xc = (x1+x2) / 2;
			String name = "Nx"+x1+"_"+x2+"y"+y1+"_"+y2;

			n = makeNode( makeGrid(x1,xc,y1,y2,1,level),
					  makeGrid(xc,x2,y1,y2,1,level),
					  0,xc*xSpacing);

			n.setName(name);
			return n;

		}  else { // split along y
			int yc = (y1+y2) / 2;

			return makeNode( makeGrid(x1,x2,y1,yc,0,level-1),
					  makeGrid(x1,x2,yc,y2,0,level-1),
					  1,yc*zSpacing);

		}

	}




	// Event handlers
	
    public boolean OnScrollbar(Scrollbar bar,Event event) 
	{

		if (bar == fracDimScr) { // got scroll 
			fracDim = (float) (((float) fracDimScr.getValue()) * 0.001f); 
			fracDimField.setText(Float.toString(fracDim));
			return true;
		} 
		if (bar == seedScr) { // got scroll 
			seed = seedScr.getValue(); 
			seedField.setText(String.valueOf(seed));
			return true;
		}
		if (bar == sigmaScr) { // got scroll 
			sigma = (float) (((float) sigmaScr.getValue()) * 0.1f); 
			sigmaField.setText(String.valueOf(sigma));
			return true;
		}
		if (bar == spacingScr) { // got scroll 
			xSpacing = zSpacing = (float) (((float) spacingScr.getValue()) * 0.1f); 
			spacingField.setText(String.valueOf(xSpacing));
			return true;
		}

		return true;
	}


    public boolean OnButton(Button button,Event event, Object what) {

		String cmd = button.getLabel();

		String s= new String();
		s = "Got Button " + button.getLabel();
	    showStatus(s);

		if (cmd == "Go") {
			OnCompute();
		}

		return true;
	}

    public boolean OnInputField(TextField field,Event event, Object what) {


		if (field == fracDimField) {
			float x = Float.valueOf((String) what).floatValue();
			if (true) {
				fracDim = x;
			    fracDimScr.setValue((int) (fracDim*10f));
				OnCompute();

			}	
			return true;
		}			
		if (field == sigmaField) {
			float x = Float.valueOf((String) what).floatValue();
			if (true) {
				sigma = x;
			    sigmaScr.setValue((int) (fracDim*1000f));
				OnCompute();

			}	
			return true;
		}			
		if (field == seedField) {
			int x = Integer.valueOf((String) what).intValue();
			if (true) {
				seed = x;
			    seedScr.setValue(seed);
				OnCompute();

			}	
			return true;
		}			

		if (field == powerField) {
			int x = Integer.valueOf((String) what).intValue();
			if (x>=1 && x <= 12) {
				power = x;
				OnCompute();

			}	
			return true;
		}			
		if (field == valleysField) {
			int x = Integer.valueOf((String) what).intValue();
			if (x>=0) {
				numValleys = x;
				OnCompute();
			}	
			return true;
		}			
		if (field == cratersField) {
			int x = Integer.valueOf((String) what).intValue();
			if (x>=0) {
				numCraters = x;
				OnCompute();
			}	
			return true;
		}			

		if (field == spacingField) {
			float x = Float.valueOf((String) what).floatValue();
			if (true) {
				if (x == 0.0f) x = 100.0f / (1<<power);
				xSpacing=zSpacing = x;
			    spacingScr.setValue((int) (xSpacing*1000f));
				OnCompute();

			}	
			return true;
		}			

		return true;
	}


	// event dispatch 
    public boolean handleEvent(Event event) {
	    if (event.target instanceof Scrollbar)
		{

			Scrollbar bar = (Scrollbar) event.target;
			return OnScrollbar(bar,event);
		}

	    return super.handleEvent(event);
	}
	
	public boolean action(Event event, Object what) {
		if (event.target instanceof Button) {
			return OnButton((Button) event.target,event,what);
		}
		else if (event.target instanceof TextField) {
			return OnInputField((TextField) event.target,event,what);
		}
	    return super.action(event,what);
	}


	// VRML helpers 

	// make Appearance with randomized material 
/*
	Node makeAppearance() {
			Node a=createNode("Appearance");
			Node m=createNode("Material");

			float c[] = new float[3];
			c[0]= (float) random1.nextGaussian(); 

			c[1]= 0.7f; c[2]= 0.4f;
			((EventInSFColor) m.getEventIn("set_diffuseColor")).setValue(c);

			c[0]= 1.0f; c[1]= 1.0f; c[2]= 1.0f;
			//((EventInSFColor) m.getEventIn("set_specularColor")).setValue(c);

			((EventInSFNode) a.getEventIn("set_material")).setValue(m);
			return a;

	}
*/

	// make appearance with unlit texture 
	Node makeAppearanceTexture(String url) {
			Node a=createNode("Appearance");
			Node t=createNode("ImageTexture");

			String urlMf[] = new String[1];
			urlMf[0] = url;
			((EventInMFString) t.getEventIn("set_url")).setValue(urlMf);


			((EventInSFNode) a.getEventIn("set_texture")).setValue(t);
			return a;

	}

	// create a shape node 
	Node makeShape(Node geometry, Node appearance) 
	{
		Node n;
		n = createNode("Shape");
		((EventInSFNode) n.getEventIn("set_appearance")).setValue(appearance);
		((EventInSFNode) n.getEventIn("set_geometry")).setValue(geometry);
		return n;

	}


	// create coordinate node 
	Node makeCoord(float p[][]) {
		Node n;
		n = createNode("Coordinate");
		((EventInMFVec3f) n.getEventIn("set_point")).setValue(p);
		return n;
	}

	// create a texCoordinate node 
	Node makeTexCoord(float p[][]) {
		Node n;
		n = createNode("TextureCoordinate");
		((EventInMFVec2f) n.getEventIn("set_point")).setValue(p);
		return n;
	}

	// create color node 
	Node makeColor(float p[][]) {
		Node n;
		n = createNode("Color");
		((EventInMFColor) n.getEventIn("set_color")).setValue(p);
		return n;
	}

	// create box 
	Node makeBox(float dx,float dy, float dz) {
		Node n;
		n = createNode("Box");
		float p[] = new float[3];
		p[0]=dx; p[1]=dy;p[2]=dz;
		((EventInSFVec3f) n.getEventIn("set_size")).setValue(p);
		return n;
	}








}

